import os
import compiler
from bike.parsing.load import load as loadast
from bike.refactor import renameFunction, renameClass, renameMethod, extractMethod
from bike.parsing.output import save as saveast
from bike.parsing.utils import fqn_rcar, fqn_rcdr
from bike.parsing.tokenhandler import coords
from bike.parsing.undo import getUndoStack, UndoStackEmptyException


# Returns a BRMContext object

# progressCallback is an optional callback which is called when brm

# wants to indicate progress. The object must be callable and takes

# the following parameters:

#  - action  (string)

#  - filename (string)

def load(path, progressCallback = None):
    context = BRMContext_impl(path, progressCallback)
    return context

# the context object public interface

class BRMContext(object):

    def load(self, path):
        """ import some more code into bicyclerepairman """

    def save(self):
        """ save the changed files out to disk """

    def setRenameMethodPromptCallback(self, callback):
        """

        sets a callback to ask the user about method refs which brm

        can't deduce the type of. The callback must be callable, and

         take the following parameters:

          - filename

          - start coords   (object with integer x and y attributes)

          - end coords   (object with integer x and y attributes)

         (start and end coords enclose the problematic method call)

        """

    def renameClass(self, fqn, newname):
        """ renames a class. fqn is the fully qualified name of the

        class from the root package"""

    def renameFunction(self, fqn, newname):
        """ see renameClass """

    def renameMethod(self, fqn, newname):
        """ see renameClass """

    def renameByCoordinates(self, filename_path, x, y, newname):
        """ an ide friendly method which renames a class/fn/method

        pointed to by the coordinates and filename"""

    def extractMethod(self, filename_path, 
                      begin_x, begin_y, 
                      end_x, end_y, 
                      methodname):
        """ extracts the region into the named method"""

    def extractFunction(self, filename_path, 
                        begin_x, begin_y, 
                        end_x, end_y, 
                        functionname):
        """ extracts the region into the named function"""

    def undo(self):
        """ undoes the last refactoring. WARNING: this is dangerous if

        the user has modified files since the last refactoring.

        Raises UndoStackEmptyException"""

class NotAPythonModuleOrPackageException: pass
class CouldntLocateASTNodeFromCoordinatesException: pass

class BRMContext_impl(BRMContext):
    def __init__(self, path, progressCallback):
        self.progressCallback = progressCallback
        self.ast = None

        # Used because some refactorings delegate back to the user.
        # this flag ensures that code isnt imported during those times
        self.readyToLoadNewCode = 1 

        self.load(path)
        getUndoStack(1)  # force new undo stack


    def load(self, path):
        if not self.readyToLoadNewCode:
            return
            
        if(not path.endswith(".py"))and os.path.isfile(path):
            raise NotAPythonModuleOrPackageException
        path = os.path.expanduser(path)
        path = os.path.abspath(path)
        class SourceVisitor:
            def visitSource(self, node):
                if node.filename == path:
                    # backup the code
                    getUndoStack().addSource(node,node.source)
                    getUndoStack().commitUndoFrame()
        if self.ast:
            compiler.walk(self.ast, SourceVisitor())
        self.ast = loadast(path, self.progressCallback, self.ast)

    def _getAST(self):
        return self.ast

    # returns a list of saved filenames

    def save(self):
        savedfiles = saveast(self.ast)
        return savedfiles

    def setRenameMethodPromptCallback(self, callback):
        self.promptUserClientCallback = callback



    # fqn = fully qualified name  (a.b.foo.bah)

    def renameClass(self, fqn, newname):
        renameClass.renameClass(self.ast, fqn, newname,
                                progressCallback = self.progressCallback)

    def renameFunction(self, fqn, newname):
        renameFunction.renameFunction(self.ast, fqn, newname,
                                      progressCallback = self.progressCallback)

    def renameMethod(self, fqn, newname):
        promptcallback = None
        if hasattr(self, "promptUserClientCallback"):
            promptcallback = self._promptUser
        renameMethod.renameMethod(self.ast, fqn, newname, 
                                  promptCallback = promptcallback, 
                                  progressCallback = self.progressCallback)

    def extractMethod(self, filename_path, 
                      begin_x, begin_y, 
                      end_x, end_y, 
                      methodname):
        ast = self.getSourceNode(filename_path)
        assert ast
        extractMethod.extractMethod(ast, coords(begin_x, begin_y), 
                                    coords(end_x, end_y), methodname)


    def extractFunction(self, filename_path, 
                      begin_x, begin_y, 
                      end_x, end_y, 
                      methodname):
        ast = self.getSourceNode(filename_path)
        assert ast
        extractMethod.extractFunction(ast, coords(begin_x, begin_y), 
                                      coords(end_x, end_y), methodname)

    def undo(self):
        getUndoStack().undo()


    def _promptUser(self, node, contextNode):
        startcoords = node.getStartCoords()
        endcoords = node.getEndCoords()
        filename = node.sourcenode.filename
        return self.promptUserClientCallback(filename, startcoords, endcoords)


    # filename_path must be absolute

    def renameByCoordinates(self, filename_path, x, y, newname):
        self.readyToLoadNewCode = 0
        try:
            node = self._locateNode(filename_path, x, y)
            if node:
                if isinstance(node, compiler.ast.Function):
                    if isinstance(node.parentScope, compiler.ast.Class):
                        self.renameMethod(node.fqn, newname)
                    else:
                        self.renameFunction(node.fqn, newname)
                if isinstance(node, compiler.ast.Class):
                    self.renameClass(node.fqn, newname)
            else:
                raise CouldntLocateASTNodeFromCoordinatesException()
        finally:
            self.readyToLoadNewCode = 1


    def getSourceNode(self, filename):
        top = self.ast
        class SourceVisitor:
            srcnode = None
            def visitSource(self, node):
                #print node.filename,filename

                if node.filename == filename:
                    self.srcnode = node
        return compiler.walk(self.ast, SourceVisitor()).srcnode

    def _writeProgress(self, action, filename):
        if self.progressCallback is not None:
            self.progressCallback(action, filename)

    def _locateNode(self, filename, x, y):
        writeProgress = self._writeProgress
        class SourceVisitor:
            def __init__(self):
                self.result = None
            def visitSource(self, node):
                #print "source",node.modulename

                if node.filename == filename:
                    #print "found"

                    writeProgress("Tokenising Source", node.filename)
                    node.tokenize()
                    self.visit(node.module)

            def visitFunction(self, node):
                startcoords = node.getStartCoords()
                #print "coords",node.name,startcoords

                if startcoords.y == y:
                    self.result = node
                    return
                for child in node.getChildNodes():
                    self.visit(child)

            def visitClass(self, node):
                startcoords = node.getStartCoords()
                if startcoords.y == y:
                    self.result = node
                    return
                for child in node.getChildNodes():
                    self.visit(child)

        result = compiler.walk(self.ast, SourceVisitor()).result
        return result

def _deducePackageOfFile(filename):
    package = ""
    dot = ""
    dir = os.path.dirname(filename)
    while dir != ""and \
          os.path.exists(os.path.join(dir, "__init__.py")):
        dir, dirname = os.path.split(dir)
        package = dirname+dot+package
        dot = "."
    return package
