import compiler
import constants
from typeinfo import *
from utils import fqn_rcar, fqn_rcdr, fqn_car, fqn_cdr

import bike.parsing.extended_ast

# parentnode is for when this is called on a sourcenode half way down
# the tree (for reloading source files)
# ast can either be a 'Top', or a 'Source' ast node
def addtypeinfo(ast,parentnode=None,**opts):
    progressCallback = None
    if opts.has_key('progressCallback'):
        progressCallback = opts['progressCallback']
    compiler.walk(ast,PassOne_ModulesClassesAndFunctions(parentnode))
    compiler.walk(ast,PassTwo_AssignmentsAndReturnsFromFunctionCalls(progressCallback))

def makeFQN(scopename,leafname):
    if leafname == "":
        return scopename
    elif scopename == "":
        return leafname
    else:
        return scopename+'.'+leafname


class PassOne_ModulesClassesAndFunctions:

    def __init__(self,parentnode):
        self.parentnode = parentnode
        pass
        
    def visitTop(self,node):
        node.fqn = ""
        node.typeinfo = TypeInfo()
        for child in node.getChildNodes():
            self.visit(child,node)

    def visitPackage(self,node,scope=None):
        if scope is None:
            scope = self.parentnode
        node.parentScope = scope
        node.fqn = makeFQN(node.parentScope.fqn,node.name)
        node.typeinfo = TypeInfo(node.parentScope)
        node.parentScope.typeinfo.addType(node.name,node)
        for child in node.getChildNodes():
            self.visit(child,node)

    def visitSource(self,node,scope=None):
        if scope is None:
            scope = self.parentnode
        node.parentScope = scope
        for c in node.getChildNodes():
            self.visit(c,scope,node)
        
    def visitModule(self,node,scope,sourcenode):
        node.parentScope = scope
        node.fqn = makeFQN(node.parentScope.fqn,node.name)
        node.sourcenode = sourcenode
        node.parentScope.typeinfo.addType(node.name,node)
        node.typeinfo = ModuleTypeInfo(node.parentScope)
        for c in node.getChildNodes():
            self.visit(c,node,sourcenode)

    def visitClass(self,node,scope,sourcenode):
        node.parentScope = scope
        node.fqn = makeFQN(node.parentScope.fqn,node.name)
        node.sourcenode = sourcenode
        node.typeinfo = ClassTypeInfo(node.parentScope)
        node.parentScope.typeinfo.addType(node.name,node)
        for child in node.getChildNodes():
            self.visit(child,node,sourcenode)

    def visitFunction(self,node,scope,sourcenode):
        node.parentScope = scope
        node.fqn = makeFQN(node.parentScope.fqn,node.name)
        node.sourcenode = sourcenode
        node.typeinfo = TypeInfo(node.parentScope)
        node.parentScope.typeinfo.addType(node.name,node)
        if len(node.argnames)!=0 and node.argnames[0] == "self":
            self._addSelfToFunction(node,node.parentScope)
        for c in node.getChildNodes():
            self.visit(c,node,sourcenode)

    def _addSelfToFunction(self,fnnode,classnode):
        fnnode.typeinfo.addType("self",bike.parsing.extended_ast.Instance(classnode))

    # all other nodes
    def __getattr__(self, name): 
        if(name.startswith("visit")):
            return self.visitNode
        else:
            raise AttributeError, name

    def visitNode(self, node, scope,sourcenode):
        if node != []:
            node.parentScope = scope
            node.sourcenode = sourcenode
        for c in node.getChildNodes():
            self.visit(c,scope,sourcenode)



class PassTwo_AssignmentsAndReturnsFromFunctionCalls:

    def __init__(self,promptCallback,calldepth=0):
        self.calldepth = calldepth
        self.promptCallback = promptCallback

    def visitPackage(self,node,scope=None):
        for child in node.getChildNodes():
            self.visit(child,node)

    def visitModule(self,node,scope=None):
        if self.promptCallback:
            self.promptCallback("Deducing Type Information",node.name)
        self.visit(node.node,node)

    def visitClass(self,node,scope=None):
        self.visit(node.code,node)
    
    def visitFunction(self,node,scope=None):
        self.visit(node.code,node)

    def visitAssign(self,assign,scope):
        if isinstance(assign.expr,compiler.ast.CallFunc):
            # store type as a closure for later (lazy) evaluation
            type = lambda : self._getTypeOfCallFunc(assign.expr,scope)   
            for assnode in assign.nodes:
                if isinstance(assnode,compiler.ast.AssName):
                    scope.typeinfo.addType(assnode.name,type)
                elif isinstance(assnode,compiler.ast.AssAttr):
                    node = scope.typeinfo.getTypeOfExpr(assnode.expr)
                    if node is not None:
                        if isinstance(node,bike.parsing.extended_ast.Instance):
                            node.getType().typeinfo.addType(assnode.attrname,type)
                        else:
                            node.typeinfo.addType(assnode.attrname,type)


    def visitReturn(self,node,scope):
        scope.typeinfo.addReturnType(lambda :self.calcReturnType(node,scope))

    def calcReturnType(self,node,scope):
        if isinstance(node.value,compiler.ast.CallFunc):
            return self._getTypeOfCallFunc(node.value,scope)
        else:
            try:
                return scope.typeinfo.getTypeOfExpr(node.value)
            except TypeNotSupportedException:
                return None


    def visitImport(self,node,scope):
        for name,alias in node.names:
            if alias is not None:
                importedname = alias
                # closure for lazy eval, just in case module changes (is reloaded)
                #   (name=name insures value rather than ref is used in closure)
                type = lambda name=name:self._getImportedType(scope,name) 
                scope.typeinfo.addType(importedname,type)
            else:
                # dont need to do anything for normal imports - the type
                # lookup will just work since typeinfo delegates to the
                # parent scope
                pass


    def visitFrom(self,node,scope):
        if node.names[0][0] == '*': # e.g. from foo import *
            scope.typeinfo.addModuleImportedWithStar(node.modname)
        else:
            for name,alias in node.names:      
                if alias is not None:
                    importedname = alias
                else:
                    importedname = name
                # closure for lazy eval, just in case module changes (is reloaded)
                #   (name=name insures value rather than ref is used in closure)
                type = lambda name=name:self._getImportedType(scope,node.modname,name)
                scope.typeinfo.addType(importedname,type)

    # returns None if cant get type
    def _getTypeOfCallFunc(self,callfunc,scope):
        try:
            #print "calling ",callfunc
            node = scope.typeinfo.getTypeOfExpr(callfunc.node)
            #print node
        except TypeNotSupportedException:
            return None
        if isinstance(node,compiler.ast.Class):
            return bike.parsing.extended_ast.Instance(node)
        elif isinstance(node,compiler.ast.Function):
            return node.typeinfo.getReturnType()
        else: 
            return None
                                       

    def _getImportedType(self,scope,modname,name=None):
        toptypeinfo = scope.sourcenode.topnode.typeinfo   # code smell?
        
        parent_package = fqn_rcdr(scope.fqn)
        if name == "*" or name == None:
            dot_name = ""
        else:
            dot_name = "."+name
        if toptypeinfo.getTypeOf_doesntThrowException(parent_package + "." +modname + dot_name):
            fqn = parent_package + "." +modname + dot_name
        elif toptypeinfo.getTypeOf_doesntThrowException(modname + dot_name):
            fqn = modname + dot_name
        else:
            return None  # can't find imported type
        return toptypeinfo.getTypeOf_doesntThrowException(fqn)        
        
