import setpath
from compiler.ast import*
import tokenhandler
import compiler
from bike.parsing.addtypeinfo import addtypeinfo
import matchTokensToAST
import tokenutils
import brmtransformer
import os
from bike.parsing.undo import getUndoStack
from bike.parsing.output import tokensToString

# dodgy hack to retrofit the token functionality to existing
# compiler.ast classes.

# n.b. not sure this works with jython

Node.__bases__ = (tokenhandler.SimpleTokenHandler, )
Name.__bases__ = (tokenhandler.NameTokenHandler, Node)
Pass.__bases__ = (tokenhandler.SimpleTokenHandler, Node)
Class.__bases__ = (tokenhandler.ClassTokenHandler, Node)
Function.__bases__ = (tokenhandler.FunctionTokenHandler, Node)
Module.__bases__ = (tokenhandler.ModuleTokenHandler, Node)
Getattr.__bases__ = (tokenhandler.GetattrTokenHandler, Node)
Import.__bases__ = (tokenhandler.ImportTokenHandler, Node)
From.__bases__ = (tokenhandler.FromTokenHandler, Node)

class NoSuchPackageException:
    pass
class PackageAlreadyExistsException:
    pass
class NoSuchModuleException:
    pass
class ModuleAlreadyExistsException:
    pass

class ExtendedNode(Node):
    def _getChildren(self):
        return self._children

class PackageContainer:
    def __init__(self, packages):
        self.packages = packages

    def getPackage(self, packagename):
        packagenames = [x.name for x in self.packages]
        if packagename not in packagenames:
            raise NoSuchPackageException()
        else:
            i = packagenames.index(packagename)
            return self.packages[i]


    def addPackage(self, package):
        packagenames = [x.name for x in self.packages]
        if package.name in packagenames:
            raise PackageAlreadyExistsException()
        else:
            package.topnode = self.topnode
            package.setParent(self)
            if self.fqn == "":
                package.setFQN(package.name)
            else:
                package.setFQN(self.fqn+"."+package.name)
            self.packages.append(package)

class SourceContainer:
    def __init__(self, modules, topnode = None):
        self.modules = modules
        # self.topnode should be set by parent

    def addSource(self, source):
        if source.modulename in[x.modulename for x in self.modules]:
            raise ModuleAlreadyExistsException()
        source.setTopNode(self.topnode)
        source.setParent(self)
        if self.fqn == "":
            source.setFQN(source.modulename)
        else:
            source.setFQN(self.fqn+"."+source.modulename)
        self.modules.append(source)


    def getSource(self, modulename):
        modnames = [x.modulename for x in self.modules]
        try:
            return self.modules[modnames.index(modulename)]
        except ValueError:
            raise NoSuchModuleException()

    def hasSource(self, modulename):
        return modulename in[x.modulename for x in self.modules]


class Top(ExtendedNode, PackageContainer, SourceContainer):
    # Always the root of the extended ast.
    nodes['top'] = 'Top'

    def __init__(self, packages = None, modules = None, pythonpath = None):

        # this is to get round an optimisation which reuses an empty list as a default arg
        #  unfortunately the client of this method may fill that list, so it's not empty
        if not packages:
            packages = []
        if not modules:
            modules = []
        if not pythonpath:
            pythonpath = []
        PackageContainer.__init__(self, packages)
        SourceContainer.__init__(self, modules)
        self.pythonpath = pythonpath
        self._children = ('top', packages, modules, pythonpath)
        self.topnode = self
        addtypeinfo(self)

    def __repr__(self):
        return "Top(%s,%s,%s)"%self._children[1:]

    def getChildNodes(self):
        nodes = []
        if self.packages is not None: nodes+=self.packages
        if self.modules is not None: nodes+=self.modules
        return tuple(nodes)




class Source(ExtendedNode):
    nodes['Source'] = 'Source'

    def __init__(self, filename, modulename, source):
        if os.path.isabs(filename):
            self.filename = filename
        else:
            self.filename = os.path.abspath(filename)
        self.modulename = modulename
        self.load(source)
        self._children = ('source', filename, modulename, self.module)

    def load(self, source):
        # tokens
        self.tokens = tokenutils.getTokenList(source)
        self._tokenized = 0

        # ast
        self.module = brmtransformer.parse(source)
        # retrofit name to module
        self.module.name = self.modulename
        self._modified = 0
        self.sourcenode = self
        self.source = source

        # COMMENT ME OUT
        #self.tokenize()


    def __repr__(self):
        return "Source(%s,%s,%s)"%self._children[1:]

    def getChildren(self):
        return self.getChildNodes()

    def getChildNodes(self):
        return tuple([self.module])

    def isModified(self):
        return self._modified

    def refreshTokens(self):
        src = self.getSourceAsString()
        self.reload(src)

    def setModified(self):
        self._modified = 1

    def setUpToDate(self):
        self._modified = 0

    def setTopNode(self, topnode):
        self.topnode = topnode

    def reload(self, source):
        self.load(source)
        addtypeinfo(self, self.parentScope)

    def setFQN(self, fqn):
        self.fqn = fqn

    def setParent(self, scope):
        self.parentScope = scope
        self.addTypeinfoToNodes()

    def addTypeinfoToNodes(self):
        addtypeinfo(self, self.parentScope)

    def tokenize(self):
        if not self._tokenized:
            matchTokensToAST.matchTokensToAST(self.module, self.getTokens())
            self._tokenized = 1

    def getTokens(self):
        # you need tokens to modify the source, so this is a good place
        # to hook into undo functionality
        getUndoStack().addSource(self,self.source)
        return self.tokens

    # same as above, but doesnt do the undo bit
    def getTokens_internal(self):
        return self.tokens

    def getSourceAsString(self):
        tokens = self.getTokens_internal()
        return tokensToString(tokens.begin(), tokens.end())

class CExtension(ExtendedNode):
    # good for anything?
    nodes['CExtension'] = 'CExtension'

    def __init__(self, path, name):
        self.path = path
        self.name = name
        self._children = ('c_extension', path, name)

    def __repr__(self):
        return "CExtension(%s,%s)"%self._children[1:]

class Package(ExtendedNode, PackageContainer, SourceContainer):
    nodes['package'] = 'Package'

    def __init__(self, path, name, topnode = None):
        if topnode is not None:
            self.topnode = topnode
        PackageContainer.__init__(self, [])
        SourceContainer.__init__(self, [])
        self.path = path
        self.name = name
        self._children = self.packages, self.modules

    def setFQN(self, fqn):
        self.fqn = fqn

    def setParent(self, scope):
        self.parentScope = scope
        addtypeinfo(self, self.parentScope)

    def getChildNodes(self):
        nodes = []
        if self.packages is not None: nodes+=self.packages
        if self.modules is not None: nodes+=self.modules
        return tuple(nodes)

    def __repr__(self):
        return "Package(%s,%s,%s,%s)"%(self.path, self.name, self.packages, self.modules)

# describes an instance of something (e.g. instance of a class)
class Instance(compiler.ast.Node):
    def __init__(self, type):
        assert type is not None
        self._type = type

    def getType(self):
        return self._type

    def __str__(self):
        return "Instance(%s)"%(self.getType())




# Local Variables:          
# mode:python               
# indent-tabs-mode: nil     
# py-indent-offset: 4   
# py-smart-indentation: nil 
# End:                      
