#!/usr/bin/env python 
import setpath
import unittest
import sys

from bike import testdata
from bike.testutils import*
from brmtransformer import parse
from addtypeinfo import*
from load import load
from matchTokensToAST import matchTokensToASTFromSrc
from compiler.ast import Getattr, Name
from bike.parsing.extended_ast import Top, Source, Instance


# adds type info to existing ast trees (when this is complete, all
# scope nodes should contain typeinfo map attributes)
class TestAddTypeInfo(unittest.TestCase):
    def setUp(self):
        self.module = "a."
        self.theclassname = "a.TheClass"
        self.foofunctionname = "a.a.foo"
        self.aclassname = "a.a"
        self.modulename = "bah"

    # template method
    def getASTForTesting(self, src):
        ast = createSourceNodeAt(Top(), src, "bah")
        addtypeinfo(ast)
        return ast

    def testAddTypeInfo_addsClassTypeToModule(self):
        ast = self.getASTForTesting(testdata.class_and_method[1:])
        modulenode = getMatchingASTNode(ast, "Module", name = self.modulename)
        classnode = getASTNode(ast, "Class")
        assert modulenode.typeinfo.getTypeOf('a') == classnode

    def testAddTypeInfo_addsInstanceRefToModule(self):
        src = testdata.ObjectCreatedInParentScopeAfterMethodReference[1:]
        ast = self.getASTForTesting(src)
        modulenode = getMatchingASTNode(ast, "Module", name = self.modulename)
        classnode = getMatchingASTNode(ast, "Class", name = "TheClass")
        assert modulenode.typeinfo.getTypeOf('a').getType() == classnode

    def testAddTypeInfo_addsInstanceRefToFunction(self):
        src = testdata.ObjectCreatedInFunctionAndUsedByCaller[1:]
        ast = self.getASTForTesting(src)
        fnnode = getASTNode(ast, "Function", 1)
        classnode = getMatchingASTNode(ast, "Class", name = "TheClass")
        assert isinstance(fnnode.typeinfo.getTypeOf('b'), Instance)
        assert fnnode.typeinfo.getTypeOf('b').getType() == classnode


    def testAddTypeInfo_addsReturnTypeToFunction(self):
        src = testdata.ObjectCreatedInFunctionAndUsedByCaller[1:]
        ast = self.getASTForTesting(src)
        fnnode = getASTNode(ast, "Function", 1)
        classnode = getMatchingASTNode(ast, "Class", name = "TheClass")
        #print fnnode.typeinfo        
        assert fnnode.typeinfo.getReturnType().getType() == classnode


    def testAddTypeInfo_deducesInstanceRefTypeWhenResultOfFunctionCall(self):
        src = testdata.ObjectCreatedInFunctionAndUsedByCaller[1:]
        ast = self.getASTForTesting(src)
        fnnode = getASTNode(ast, "Function")
        classnode = getMatchingASTNode(ast, "Class", name = "TheClass")
        assert fnnode.typeinfo.getTypeOf('b') == classnode

    def testAddTypeInfo_addsFunctionTypeToModule(self):
        src = testdata.ObjectCreatedInFunctionAndUsedByCaller[1:]
        ast = self.getASTForTesting(src)
        modulenode = getMatchingASTNode(ast, "Module", name = self.modulename)
        fnnode = getASTNode(ast, "Function", 1)
        assert modulenode.typeinfo.getTypeOf('foo') == fnnode

    def testAddTypeInfo_deducesInstanceRefTypeWhenResultOfFunctionCall(self):
        src = testdata.ObjectCreatedInFunctionAndUsedByCaller[1:]
        ast = self.getASTForTesting(src)
        modulenode = getMatchingASTNode(ast, "Module", name = self.modulename)
        classnode = getMatchingASTNode(ast, "Class", name = "TheClass")
        assert modulenode.typeinfo.getTypeOf('a').getType() == classnode

    def testAddTypeInfo_deducesInstanceRefTypeWhenReturnedByFunctionCall(self):
        src = testdata.ObjectReturnedByFunctionAndUsedByCaller[1:]
        ast = self.getASTForTesting(src)
        modulenode = getMatchingASTNode(ast, "Module", name = self.modulename)
        classnode = getMatchingASTNode(ast, "Class", name = "TheClass")
        assert modulenode.typeinfo.getTypeOf('a').getType() == classnode

    def testAddTypeInfo_deducesInstanceRefTypeWhenReturnedByFnCallFromInsideAnotherFn(self):
        src = testdata.ObjectReturnedByFunctionAndUsedByCallerInOtherFunction[1:]
        ast = self.getASTForTesting(src)
        functionnode = getASTNode(ast, "Function", 2)
        classnode = getMatchingASTNode(ast, "Class", name = "TheClass")
        assert functionnode.typeinfo.getTypeOf('a').getType() == classnode

    def testAddTypeInfo_deducesInstanceRefTypeWhenReturnedByFnCallFromInsidePreviousFn(self):
        src = testdata.ObjectReturnedByFunctionAndUsedByCallerInPreviousFunction[1:]
        ast = self.getASTForTesting(src)
        functionnode = getASTNode(ast, "Function", 2)
        classnode = getMatchingASTNode(ast, "Class", name = "TheClass")
        assert functionnode.typeinfo.getTypeOf('a').getType() == classnode


    def testAddTypeInfo_handlesImportedTypesItCantFindBySettingTypeToNoneType(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceWithImport[1:]
        ast = self.getASTForTesting(src)
        functionnode = getASTNode(ast, "Function", 0)
        assert functionnode.typeinfo.getTypeOf('a') == None

    def testAddTypeInfo_doesntStackOverflowOnRecursiveFunctions(self):
        src = testdata.RecursiveFunctionDef[1:]
        ast = self.getASTForTesting(src)

    def testAddTypeInfo_knowsTypeOfSelf(self):
        src = testdata.ClassWithRefToOwnMethod
        ast = self.getASTForTesting(src)
        functionnode = getASTNode(ast, "Function", 0)
        classnode = getASTNode(ast, "Class", 0)
        assert functionnode.typeinfo.getTypeOf("self").getType() == classnode


    def testAddTypeInfo_doesntStackOverflowOnIndirectRecursiveFunction(self):
        src = testdata.IndirectRecursiveFunctionCall
        ast = self.getASTForTesting(src)

    def testAddTypeInfo_doesntLoopForeverOnRecursiveFunction(self):
        src = testdata.RecursiveFunctionCall
        ast = self.getASTForTesting(src)

    def testAddTypeInfo_functionTypeinfoContainsNestedClasses(self):
        src = testdata.ClassNestedInFunction
        ast = self.getASTForTesting(src)
        functionnode = getASTNode(ast, "Function", 0)
        classnode = getASTNode(ast, "Class", 0)
        assert functionnode.typeinfo.getTypeOf('TheClass') == classnode

class TestAddTypeInfoWorksThroughLoad(TestAddTypeInfo):
    def setUp(self):
        self.module = tmpmodule+"."
        self.theclassname = tmpmodule+".TheClass"
        self.foofunctionname = tmpmodule+".a.foo"
        self.aclassname = tmpmodule+".a"
        self.modulename = tmpmodule

    # template method
    def getASTForTesting(self, src):
        writeTmpTestFile(src)
        ast = load(tmpfile)
        deleteTmpTestFile()
        return ast




class TestAddTypeInfoWorksWithPackageStructure(TestAddTypeInfoWorksThroughLoad):
    def setUp(self):
        self.module = "a.b.bah."
        self.theclassname = "a.b.bah.TheClass"
        self.foofunctionname = "a.b.bah.a.foo"
        self.aclassname = "a.b.bah.a"
        self.modulename = "bah"

    # template method
    def getASTForTesting(self, src):
        createPackageStructure("", src)
        ast = load(pkgstructureBasedir)
        removePackageStructure()
        return ast


class TestAddTypeInfoWithTheClassInNestedScope(unittest.TestCase):

    # template method
    def getASTWithTestClassInSubPackage(self, src):
        ast = Top()
        createSourceNodeAt(ast, src, "a.foo")
        createSourceNodeAt(ast, testdata.TheClass, "a.b.bah")
        addtypeinfo(ast)
        return ast


class CreateASTHelper:
    def getASTForTesting(self, src):
        ast = createSourceNodeAt(Top(), src, "foo")
        addtypeinfo(ast)
        return ast

    def getASTWithTestClassInSubPackage(self, src):
        ast = Top([], [], [])
        createSourceNodeAt(ast, src, "a.foo")
        createSourceNodeAt(ast, testdata.TheClass, "a.b.bah")
        return ast

    def getASTWithTestClassInSeperatePackage(self, src):
        ast = Top([], [], [])
        createSourceNodeAt(ast, src, "a.foo")
        createSourceNodeAt(ast, testdata.TheClass, "b.bah")
        return ast


class TestAddTypeInfoHandlesImports(unittest.TestCase, CreateASTHelper):

    def test_importsClassReferenceFromSubPackageIntoModule(self):
        ast = self.getASTWithTestClassInSubPackage(
        testdata.functionContaingObjectCreationAndMethodReferenceWithImport)
        modast = getASTNode(ast, "Module", 1)
        TheClass = getASTNode(ast, "Class", 0)
        assert modast.typeinfo.getTypeOfExpr(Getattr(Getattr(Name('b'), 'bah'), 'TheClass')) == TheClass

    def test_importsClassReferenceFromSeperatePackageIntoModule(self):
        ast = self.getASTWithTestClassInSeperatePackage(
        testdata.functionContaingObjectCreationAndMethodReferenceWithImport)
        modast = getASTNode(ast, "Module", 0)
        TheClass = getASTNode(ast, "Class", 0)
        assert modast.typeinfo.getTypeOfExpr(Getattr(Getattr(Name('b'), 'bah'), 'TheClass')) == TheClass


    def test_importsPackageReferenceFromSubPackageIntoModule(self):
        src = testdata.ClassWithNestedInheritanceSpec
        ast = self.getASTWithTestClassInSubPackage(src)
        packageast = getASTNode(ast, "Package", 1)
        modast = getASTNode(ast, "Module", 1)
        assert modast.typeinfo.getTypeOfExpr(Name("b")) == packageast



    def test_importsClassReferenceUsingImportAs(self):
        ast = self.getASTWithTestClassInSeperatePackage(
        testdata.functionContaingObjectCreationAndMethodReferenceWithImportAs)
        modast = getASTNode(ast, "Module", 0)
        TheClass = getASTNode(ast, "Class", 0)
        assert modast.typeinfo.getTypeOfExpr(Name("MyClass")) == TheClass

    def test_importsClassReferenceFromSubPackageIntoModuleUsingFromStatement(self):
        ast = self.getASTWithTestClassInSubPackage(
        testdata.ObjectCreationAndMethodReference)
        modast = getASTNode(ast, "Module", 1)
        TheClass = getASTNode(ast, "Class", 0)
        assert modast.typeinfo.getTypeOfExpr(Name("TheClass")) == TheClass

    def test_importsClassReferenceFromSeperatePackageIntoModuleUsingFromStatement(self):
        ast = self.getASTWithTestClassInSeperatePackage(
        testdata.ObjectCreationAndMethodReference)
        modast = getASTNode(ast, "Module", 0)
        TheClass = getASTNode(ast, "Class", 0)
        assert modast.typeinfo.getTypeOfExpr(Name("TheClass")) == TheClass


    def test_importsClassReferenceUsing_From_Import_As(self):
        ast = self.getASTWithTestClassInSeperatePackage(
        testdata.functionContaingObjectCreationAndMethodReferenceWithImportFromAs)
        modast = getASTNode(ast, "Module", 0)
        TheClass = getASTNode(ast, "Class", 0)
        assert modast.typeinfo.getTypeOfExpr(Name("MyClass")) == TheClass


class TestGetTypeOfExpr(unittest.TestCase, CreateASTHelper):
    def getASTForTesting(self, src):
        ast = Top()
        ast.addSource(Source("a", "a", src))
        return ast

    def test_worksWithName(self):
        src = testdata.class_and_method
        ast = self.getASTForTesting(src)
        modast = getASTNode(ast, "Module")
        AClass = getASTNode(ast, "Class")
        assert modast.typeinfo.getTypeOfExpr(compiler.ast.Name("a")) == AClass

    def test_returnsClassDeclaredInParentScope(self):
        src = testdata.two_classes_with_same_named_method
        ast = self.getASTForTesting(src)
        modast = getASTNode(ast, "Function", 1)
        AClass = getASTNode(ast, "Class")
        assert modast.typeinfo.getTypeOfExpr(compiler.ast.Name("a")) == AClass

    def test_evaluatesGetAttrExpression(self):
        src = testdata.ClassWithNestedInheritanceSpec
        ast = self.getASTWithTestClassInSubPackage(src)
        baseClass = getASTNode(ast, "Class", 0)
        classast = getASTNode(ast, "Class", 2)
        assert classast.typeinfo.getTypeOfExpr(classast.bases[0]) == baseClass

    def test_evaluatesNestedGetAttrExpression(self):
        src = testdata.ClassInstanceIsAttributeOfSelf
        ast = self.getASTForTesting(src)
        classast = getMatchingASTNode(ast, "Class", name = 'TheClass')
        fnast = getMatchingASTNode(ast, "Function", name = 'anotherFn')
        getattrexpr = getASTNode(ast, "Getattr", 1)
        assert fnast.typeinfo.getTypeOfExpr(getattrexpr).getType() == classast

    def test_evaluatesInstanceOfClassImportedViaFromFooImportStar(self):
        src = testdata.FunctionContainingMethodReferenceWithFromImportStar
        ast = self.getASTWithTestClassInSubPackage(src)
        classnode = getMatchingASTNode(ast, "Class", name = "TheClass")
        modast = getASTNode(ast, "Module", 1)
        assert isinstance(modast.typeinfo.getTypeOfExpr(Name('a')), Instance)
        assert modast.typeinfo.getTypeOfExpr(Name('a')).getType() == classnode

    # e.g.  a = a.b.foo.TheClass()
    def test_evaluatesInstanceWhenCreatedFromFullyQualifiedClassDecl(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceWithImportOfNestedClass
        ast = self.getASTWithTestClassInSubPackage(src)
        fnast = getMatchingASTNode(ast, "Function", fqn = "a.foo.foo")
        classnode = getMatchingASTNode(ast, "Class", name = "TheClass")
        assert fnast.typeinfo.getTypeOfExpr(Name('a')).getType() == classnode

    def test_doesntbarfWhenCallMadeOnInstanceReturnedFromFnCall(self):
        src = testdata.FunctionCallReturnsInstance
        ast = self.getASTWithTestClassInSubPackage(src)

    def test_doesntbarfLookingUpARecursiveCall(self):
        src = "typeof = typeof()"
        ast = self.getASTWithTestClassInSubPackage(src)
        nameast = getASTNode(ast, "Name")
        assert nameast.sourcenode.module.typeinfo.getTypeOfExpr(nameast) == None

class TestAddSourceAttributeToModuleNodes(unittest.TestCase):
    def test_addsSourceToModuleNode(self):
        ast = Top()
        createSourceNodeAt(ast, testdata.class_and_method, "a.foo")
        addtypeinfo(ast)
        s = getASTNode(ast, "Source")
        assert s.module.sourcenode == s

    def test_addsSourceToModuleChildNodes(self):
        ast = Top()
        createSourceNodeAt(ast, testdata.class_and_method, "a.foo")
        addtypeinfo(ast)
        s = getASTNode(ast, "Source")
        fn = getMatchingASTNode(s, "Function", name = "foo")
        assert fn.sourcenode == s


class TestAddFqn(unittest.TestCase, CreateASTHelper):
    def test_parsesSourceWithoutFailing(self):
        src = testdata.source_with_all_ast_nodes
        ast = self.getASTForTesting(src)

    def test_addsFqnToSourceNodesCorrectly(self):
        ast = Top()
        createSourceNodeAt(ast, testdata.ObjectCreationAndMethodReference, "a.foo")
        createSourceNodeAt(ast, testdata.TheClass, "a.b.bah")
        addtypeinfo(ast)
        assert getMatchingASTNode(ast, "Class", name = "TheClass").fqn == "a.b.bah.TheClass"



if __name__ == "__main__":
    unittest.main()
