#!/usr/bin/env python 
import setpath
import tokenhandler
import tokenutils
import compiler.ast

from bike.testutils import getASTNode, getMatchingASTNode, createSourceNodeAt
from bike import testdata
from brmtransformer import parse
import unittest
from bike.parsing.extended_ast import *

class PackageContainerTests:
    def testGetPackage_throwsNoSuchPackageExceptionIfPackageDoesntExist(self):
        self.target.packages = []
        try:
            self.target.getPackage("doesntexist")
            self.fail("should have thrown exception")
        except NoSuchPackageException:
            pass
            
    def testGetPackage_returnsNamedPackage(self):
        self.target.packages = [Package("foo","foo","")]
        assert self.target.getPackage("foo").name == "foo"

    def testAddPackge_raisesPackageAlreadyExistsExceptionIfPackageAlreadyExists(self):
        self.target.addPackage(Package("foo","foo",""))
        try:
            self.target.addPackage(Package("foo","foo",""))
            self.fail("should have thrown exception")
        except PackageAlreadyExistsException:
            pass

    def testAddPackage_addsThePackage(self):
        p = Package("foo","foo")
        self.target.addPackage(p)
        assert self.target.getPackage("foo") == p

        

class SourceContainerTests:
    def testGetSource_throwsNoSuchModuleExceptionIfSourceDoesntExist(self):
        self.target.packages = []
        try:
            self.target.getSource("doesntexist")
            self.fail("should have thrown exception")
        except NoSuchModuleException:
            pass
            
    def testGetSource_returnsNamedSource(self):
        self.target.modules = [Source("foo","foo","hello")]
        assert self.target.getSource("foo").modulename == "foo"

    def testAddPackge_raisesModuleAlreadyExistsExceptionIfSourceAlreadyExists(self):
        self.target.addSource(Source("foo","foo","hello"))
        try:
            self.target.addSource(Source("foo","foo","hello"))
            self.fail("should have thrown exception")
        except ModuleAlreadyExistsException:
            pass

    def testAddPackage_addsThePackage(self):
        p = Package("foo","foo",[],[])
        self.target.addPackage(p)
        assert self.target.getPackage("foo") == p


class TestTop(unittest.TestCase, PackageContainerTests):
    def setUp(self):
        self.target = Top()

    def test_addPackage_setsTheFQNAndPackageScope(self):
        p = Package("bah","bah")
        self.target.addPackage(p)
        assert self.target.getPackage("bah").fqn == "bah"
        assert self.target.getPackage("bah").parentScope == self.target

    def test_addSource_setsTheFQNAndPackageScope(self):
        s = Source("bah.py","bah","class TheClass: pass")
        self.target.addSource(s)
        assert s.fqn == "bah"
        assert s.module.fqn == "bah"
        assert s.module.parentScope == self.target
        assert getASTNode(self.target,"Class").fqn == "bah.TheClass"
        assert getASTNode(self.target,"Class").parentScope == s.module


class TestPackage(unittest.TestCase, PackageContainerTests,SourceContainerTests):
    def setUp(self):
        top = Top()
        self.target = Package("foo","foo",top)
        top.addPackage(self.target)

    def test_addPackage_setsTheFQNAndPackageScope(self):
        p = Package("bah","bah")
        self.target.addPackage(p)
        assert self.target.getPackage("bah").fqn == "foo.bah"
        assert self.target.getPackage("bah").parentScope == self.target

    def test_addSource_setsTheFQNAndPackageScope(self):
        s = Source("bah.py","bah","class TheClass: pass")
        self.target.addSource(s)
        assert s.fqn == "foo.bah"
        assert s.module.fqn == "foo.bah"
        assert s.module.parentScope == self.target
        assert getASTNode(self.target,"Class").fqn == "foo.bah.TheClass"
        assert getASTNode(self.target,"Class").parentScope == s.module



class TestSource(unittest.TestCase):
    def testCreatesAbsolutePath(self):
        s = Source("foo.py","foo","")
        assert s.filename == os.path.abspath("foo.py")


        
        

class TestSourceReload(unittest.TestCase):
    def test_addsFqnToReloadedNodes(self):
        src = testdata.class_and_method
        ast = createSourceNodeAt(Top(),"pass","a.foo")
        #print ast.getPackage("a").fqn
        #s.reload
        

if __name__ == "__main__":
    unittest.main()
