#!/usr/bin/env python 
import setpath
import tokenhandler
import tokenutils
from bike.parsing.extended_ast import Top
import compiler.ast

import unittest
from tokenutils import getTokenList, getTokenVector
from test_brmtransformer import getASTNodeFromSrc
from bike.testutils import*
from bike import testdata
import output
from doublelinkedlist import region

def assertTokensAreSame(t1begin, t1end, tokens):
    it = t1begin.clone()
    pos = 0
    while it != t1end:
        assert it.deref() == tokens[pos]
        it.preincr()
        pos+=1


class TestClassTokenHandler(unittest.TestCase):

    def testRename_renamesTheClassAST(self):
        ast = createSourceNodeAt(Top(), testdata.class_and_method, "a")
        addtypeinfo(ast)
        ast = getASTNode(ast, "Class")
        ast.rename("bah")
        assert ast.name == "bah"



class TestFunctionTokenHandler(unittest.TestCase):

    def testGetStartCoords_returnsCorrectCoordsForASTOnFirstLine(self):
        src = "def foo(self): pass"
        ast = createSourceNodeAt(Top(), src, "a")
        node = getASTNode(ast, "Function")
        assert node.getStartCoords().x == 0
        assert node.getStartCoords().y == 1
        assert node.getEndCoords().x == 19
        assert node.getEndCoords().y == 1


    def testGetNodeTokens_returnsCorrectTokensForOneLineFunction(self):
        src = "def foo(self): pass"
        ast = createSourceNodeAt(Top(), src, "a")
        node = getASTNode(ast, "Function")
        begin,end = node.getNodeTokenIterators()
        assertTokensAreSame(begin,end,
                            getTokenVector(src)[:-1])


    def testGetNodeTokens_doesntBarfOnNoneAsDefaultArgToFuncition(self):
        src = testdata.method_with_NoneAsDefaultArg
        ast = createSourceNodeAt(Top(), src, "a")
        node = getASTNode(ast, "Function")
        node.getNodeTokenIterators()  # bang!


    def testGetCoords_returnsCorrectCoordsForASTNodeNotOnFirstLine(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceWithImportAs[1:]
        ast = createSourceNodeAt(Top(), src, "a")
        g = getASTNode(ast, "Function")
        assert g.getStartCoords().x == 0
        assert g.getStartCoords().y == 3
        assert g.getEndCoords().x == 0
        assert g.getEndCoords().y == 6


    def testRename_renamesTheFunction(self):
        ast = createSourceNodeAt(Top(), testdata.Function, "a")
        addtypeinfo(ast)
        ast = getASTNode(ast, "Function")
        ast.rename("bah")
        assert ast.name == "bah"

    def testRename_updatesTheTokens(self):
        ast = createSourceNodeAt(Top(), testdata.Function, "a")
        addtypeinfo(ast)
        ast = getASTNode(ast, "Function")
        ast.rename("bah")
        begin, end = ast.getTokenIterators()
        it = begin.clone()
        assert it.preincr().deref().name == "bah"

    def testRename_renamesTheTypeinfoFQN(self):
        ast = createSourceNodeAt(Top(), testdata.class_and_method, "a")
        addtypeinfo(ast)
        ast = getASTNode(ast, "Function")
        ast.rename("bah")
        assert ast.fqn == "a.a.bah"


    def testRename_updatesTheTokensIfPreceedingComments(self):
        ast = createSourceNodeAt(Top(), testdata.function_withPreceedingComment, "a")
        addtypeinfo(ast)
        ast = getASTNode(ast, "Function")
        ast.rename("bah")
        begin, end = ast.getTokenIterators()
        it = begin.clone()
        it.preincr() # move to function name tokens
        assert it.deref().name == "bah"

    def testRename_setsSourceToModified(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceWithImport
        ast = createSourceNodeAt(Top(), src, "a")
        addtypeinfo(ast)
        f = getASTNode(ast, "Function")
        s = getASTNode(ast, "Source")
        f.rename("newName")
        assert s.isModified()



    def testGetNodeTokens_returnsTokensStartingWithDefToken(self):
        ast = getASTNodeFromSrc(testdata.a_class_with_two_methods, "Function", 1)
        begin,end = ast.getNodeTokenIterators()
        assert begin.deref().name == "def"

class TestGetattrTokenHandler(unittest.TestCase):
    def testRenameAttr_renamesTheAttributeNameInAST(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceWithImport
        ast = createSourceNodeAt(Top(), src, "a")
        g = getASTNode(ast, "Getattr")
        g.renameAttr("newName")
        assert g.attrname == "newName"

    def testRenameAttr_renamesTheAttributeNameInTokenlist(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceWithImport[1:]
        ast = createSourceNodeAt(Top(), src, "a")
        g = getASTNode(ast, "Getattr")
        g.renameAttr("newName")
        assert output.astToString(g) == "b.bah.newName"

    def testRenameAttr_renamesTheAttributeNameInTokenlistWhenTokenNotSecondTokenOfList(self):
        src = testdata.RefCalledOnClassInstanceCreation
        ast = createSourceNodeAt(Top(), src, "a")
        g = getASTNode(ast, "Getattr")
        g.renameAttr("newName")
        self.assertEqual(output.astToString(g),"TheClass().newName")

    def testRenameAttr_setsSourceToModified(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceWithImport
        ast = createSourceNodeAt(Top(), src, "a")
        g = getASTNode(ast, "Getattr")
        s = getASTNode(ast, "Source")
        g.renameAttr("newName")
        assert s.isModified()

    def testGetStartCoords_returnsCorrectCoordsForASTOnFirstLine(self):
        src = "a.b"
        ast = createSourceNodeAt(Top(), src, "a")
        g = getASTNode(ast, "Getattr")
        assert g.getStartCoords().x == 0
        assert g.getStartCoords().y == 1
        assert g.getEndCoords().x == 3
        assert g.getEndCoords().y == 1

    def testGetCoords_returnsCorrectCoordsForASTNodeNotOnFirstLine(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceWithImportAs[1:]
        ast = createSourceNodeAt(Top(), src, "a")
        g = getASTNode(ast, "Getattr")
        assert g.getStartCoords().x == 4
        assert g.getStartCoords().y == 5
        assert g.getEndCoords().x == 15
        assert g.getEndCoords().y == 5


class TestImportTokenHandler(unittest.TestCase):
    def test_renamesTokensWhenFirstImportHasSameClassNameAsSecond(self):
        src = "import bah.foo, baz.foo"
        ast = createSourceNodeAt(Top(), src, "a")
        f = getASTNode(ast, "Import")
        f.rename(1, "bobbins")
        assert output.astToString(f) == "import bah.foo, baz.bobbins"


    def test_renamesClassNameTokenWhenClassIsSameNameAsModule(self):
        src = "import bah.bah"
        ast = createSourceNodeAt(Top(), src, "a")
        f = getASTNode(ast, "Import")
        f.rename(0, "bobbins")
        assert output.astToString(f) == "import bah.bobbins"

    def test_renamesClassNameTokenWhenCommentAtEndOfImport(self):
        src = "import bah.bah  # comment"
        ast = createSourceNodeAt(Top(), src, "a")
        f = getASTNode(ast, "Import")
        f.rename(0, "bobbins")
        m = getASTNode(ast, "Module")
        self.assertEqual(output.astToString(m),"import bah.bobbins  # comment")



class TestFromTokenHandler(unittest.TestCase):
    def test_renamesTokensWhenModuleIsSameNameAsImportedClass(self):
        src = "from baz import foo, baz"
        ast = createSourceNodeAt(Top(), src, "a")
        f = getASTNode(ast, "From")
        f.rename(1, "bah")
        assert output.astToString(f) == "from baz import foo, bah"

if __name__ == "__main__":
    unittest.main()
