# Utilities to handle matching of AST nodes and parse trees to the
# scanner tokens they represent. Useful because BRM needs to convert
# AST nodes back to source, and this is easier to do if you have the
# corresponding scanner tokens.

import tokenize
#import token
import types
import StringIO

from doublelinkedlist import DoubleLinkedList, region
from vector import Vector

def getTokenList(str):
    return TokenList().processString(str)

def getTokenListFromFile(path):
    return TokenList().processFile(path)

def getTokenVector(str):
    return TokenVector().processString(str)


def getTokenConverter(str):
    tlist = getTokenList(str)
    return TokenConverter(tlist.begin(), tlist.end())

def createToken(type, name):
    return TokenList.token(type, name, (0, 0), (0, 0), 0)

def _peek_prev(it):
    it.predecr()
    item = it.deref()
    it.preincr()
    return item

# produces a list of tokens
#
# N.B. this class does some token reordering to ensure that comments
# immediately before a block of code are placed after the previous
# block DEDENT token
class TokenList:

    def __init__(self):
        self.list = DoubleLinkedList()

    def processString(self,str):        
        tokenize.tokenize(StringIO.StringIO(str).readline,self.process_token)
        self.runTransformers()
        return self.list

    def processFile(self,path):
        tokenize.tokenize(open(path).readline,self.process_token)
        self.runTransformers()
        return self.list

    def runTransformers(self):
        self.putDedentTokenBeforeCommentsDescribingNextBlock()
        self.putIndentTokenBeforeCommentsDescribingBlock()
        pass

    def process_token(self,ttype,tname,tstart,tend,tline):
        t = TokenList.token(ttype,tname,tstart,tend,tline)
        self.list.append(t)


    def putDedentTokenBeforeCommentsDescribingNextBlock(self):        
        def indexOfPreviousNLCommentBlock(it):            
            j=it.clone()
            while _peek_prev(j).type == tokenize.COMMENT: j.predecr()
            while _peek_prev(j).type == tokenize.NL: j.predecr()
            if j.deref().type == tokenize.NL:
                return j
            else: return it

        #it = iter(self.list)

        it = iter(self.list)
        while it != self.list.end():
            item = it.deref()
            if item.type == tokenize.DEDENT:
                j = it.clone()
                while indexOfPreviousNLCommentBlock(j) != j:
                    j = indexOfPreviousNLCommentBlock(j)

                if it != j:
                    # move the dedent to before the comment
                    dedenttok = it.deref()
                    self.list.delete(it)
                    # fix up the line numbers
                    lineno = j.deref().start[0]
                    dedenttok.start = (lineno,dedenttok.start[1])
                    dedenttok.end = (lineno,dedenttok.end[1])
                    self.list.insert(j,dedenttok)
            it.preincr()

    # ensure comments describing block are indented at the same level
    def putIndentTokenBeforeCommentsDescribingBlock(self):
        def getIndexOfTokenAfterPreviousNewline(i):
            while _peek_prev(i).type != tokenize.NEWLINE:
                i.predecr()
            return i
            
        it = iter(self.list)
        while it != self.list.end():
            item = it.deref()
            if item.type == tokenize.INDENT:
                j = it.clone()
                j = getIndexOfTokenAfterPreviousNewline(j)
                if j != it:                    
                    indenttok = it.deref()
                    self.list.delete(it)
                    self.list.insert(j,indenttok)
            it.preincr()
    
    class token:
        def __init__(self,type,name,start,end,line):
            self.type = type
            self.name = name
            self.start= start
            self.end = end
            self.line = line

        def __eq__(self,other):
            if self.type != other.type:
                return 0
            if (self.type == tokenize.NAME or self.type == tokenize.OP) and \
               (self.name != other.name):
                return 0
            return 1

        def __ne__(self,other):
            return  not self.__eq__(other)

        def __repr__(self):
            if self.tokenDoesntRequireName():
                name = ""
            else: name = `self.name`
            return tokenize.tok_name[self.type]+name+`self.start`

        
        def tokenDoesntRequireName(self):
            return self.type == tokenize.NEWLINE\
                      or self.type == tokenize.INDENT \
                      or self.type == tokenize.DEDENT \
                      or self.type == tokenize.NL
                

# produces a vector of tokens
class TokenVector(TokenList):
    def __init__(self):
        self.list = Vector()
    

# iterates through a list of tokens, providing utilities to match
# sections of parse tree to the corresponding scanner tokens in order.
class TokenConverter:

    def __init__(self,begin,end):
        #self.tokens = DoubleLinkedList(tokenlist)
        self._begin = begin
        self._end = end
        self._it = self.begin().clone()

    def begin(self):
        return self._begin

    def end(self):
        return self._end

    def pos(self):
        return self._it
        
    def peekNextToken(self):
        if self._it == self.tokens.end():
            return 0
        else:
            return self._it.deref()

    def popToken(self):
        #self._it.pos +=1
        token = self._it.deref()
        self._it.preincr()
        return token



    def getTokensCorrespondingToParseTree(self,parsetree):
        #print "getTokensCorrespondingToParseTree",parsetree
        it = self._it.clone()
        #tokennodes = self._getTokenNodesExceptNewlines(parsetree)
        tokennodes = self._getTokenNodes(parsetree)
        #print "getTokensCorrespondingToParseTree"
        #print "parsetree", parsetree
        #print "tokennodes:",tokennodes
        #print "tokens:",region(it,self._end)
        assert it != self.end()
        endit = self._nodesMatchTokens(tokennodes,it)
        while not endit:
            it.preincr()
            #print it.deref()
            assert it != self.end()
            endit = self._nodesMatchTokens(tokennodes,it)
        self._it = endit
        return region(it,endit)
        

    def _nodesMatchTokens(self,nodes,it):
        startingit = it.clone()
        it = it.clone()
        #print it.deref()
        nodes = Vector(nodes)
        nodeit = nodes.begin().clone()
        while nodeit != nodes.end():
            node = nodeit.deref()
        
            if self._nodeMatchesToken(node[0],node[1],it.deref()):
                it.preincr()
            else:
                # parsetree nodes dont include comments
                if self._commentInMiddleOfStatement(it,startingit,node):
                    it.preincr()
                    nodeit.predecr()  # neutralize the preincr later

                elif self.tokenSaysEndMarker_nodeSaysNewline(it.deref(),
                                                             node[0]):
                    pass  # don't include the ENDMARKER
                          #(i.e. it.end = ENDMARKER)
                
                # NLs can appear in method calls without
                # '\'. These arent captured in a parsetree so be
                # ready for the superflous NL token
                elif self._newlineInMiddleOfStatement(it,startingit,node):
                    it.preincr()
                    #it.preincr()
                    nodeit.predecr()  # neutralize the preincr later
                else:
                    #print "failed on",it.deref(),nodeit.deref()
                    return 0
            nodeit.preincr()
        return it

    def _newlineInMiddleOfStatement(self, it, startingit,node):
        it = it.clone()
        if it != startingit and \
           it.deref().type == tokenize.NL:
            #if self._nodeMatchesToken(node[0],node[1],it.preincr().deref()):
            return 1
        return 0

    def _commentInMiddleOfStatement(self, it, startingit,node):
        return it != startingit and \
               it.deref().type == tokenize.COMMENT

    # returns a list of nodes that correspond to tokens
    def _getTokenNodes(self,node):

        def getTokenNodesRecursive(node,tokennodes):
            if type(node) == types.ListType or type(node) == types.TupleType:
                if self._nodeIsToken(node):
                    tokennodes.append(node)
                else:
                    for i in node:
                        getTokenNodesRecursive(i,tokennodes)
        tokennodes = []
        getTokenNodesRecursive(node,tokennodes)
        return tokennodes
        

    # returns a list of nodes that correspond to tokens
    # (but don't include newlines or endmarkers)
    def _getTokenNodesExceptNewlines(self,node):
        def getTokenNodesRecursive(node,tokennodes):
            if type(node) == types.ListType or type(node) == types.TupleType:
                if self._nodeIsToken(node) and \
                   node[0] != tokenize.NEWLINE and \
                   node[0] != tokenize.ENDMARKER:
                    tokennodes.append(node)
                else:
                    for i in node:
                        getTokenNodesRecursive(i,tokennodes)
        tokennodes = []
        getTokenNodesRecursive(node,tokennodes)
        return tokennodes

                    
    def _nodeIsToken(self,node):
        # possible token?
        if len(node) == 3 and \
           type(node[0])==types.IntType and \
           type(node[1]) == types.StringType:
            try:
                a = tokenize.tok_name[node[0]]
                return 1   # if above doesn't fail
            except:
                return 0
        return 0
        
    def _nodeMatchesToken(self,nodetype,nodename,token):
        assert not (isinstance(token,str) and token == "endmarker")
        if nodetype == tokenize.NAME and token.type == tokenize.NAME and \
           nodename == token.name:
            return 1
        if token.type == tokenize.OP and nodename == token.name:
            return 1
        if token.type != tokenize.OP and token.type != tokenize.NAME and \
           token.type == nodetype:
            return 1
        return 0


    # happens when source is all on one line -
    # parstree says newline, tokenstream says endmarker
    def tokenSaysEndMarker_nodeSaysNewline(self,token,nodetype):
        return token.type == tokenize.ENDMARKER and \
               nodetype == tokenize.NEWLINE
