#!/usr/bin/env python
import setpath
import unittest
from renameMethod import*
import compiler

from bike.parsing.addtypeinfo import addtypeinfo

from bike.parsing.brmtransformer import parse
from bike.parsing.output import astToString
from bike import testdata

from bike.testutils import*

from bike.parsing.load import load
#from bike.ui import test_bikecli

# Generic tests. These tests are designed to be run in the context of a ui
#  and in a package hierarchy structure
class RenameMethodTests:

    # template method
    def runTarget(self, src, classfqn, methodname, newname):
        # see concrete subclasses for implementation
        # klass is the fqn of the class (without modulename or path)
        # method is name of the method
        # newname is the new name
        pass

    def test_renamesTheMethod(self):
        src = testdata.two_classes_with_same_named_method[1:]
        ast = self.runTarget(src, "a.foo", "c")
        assert getASTNode(ast, "Function", 0).name == "c"

    def test_doesntRenameMethodOfSameNameOnOtherClasses(self):
        src = testdata.two_classes_with_same_named_method[1:]
        ast = self.runTarget(src, "b.foo", "c")
        assert getASTNode(ast, "Function", 0).name == "foo"
        assert getASTNode(ast, "Function", 1).name == "c"

    def test_doesntRenameOtherMethodsOfSameClass(self):
        src = testdata.a_class_with_two_methods[1:]
        ast = self.runTarget(src, "a.bah", "c")
        assert getASTNode(ast, "Function", 0).name == "foo"
        assert getASTNode(ast, "Function", 1).name == "c"

    def test_renamesMethodWhenClassNestedInFunction(self):
        src = testdata.ClassNestedInFunction
        ast = self.runTarget(src, "theFunction.TheClass.theMethod", "newName")
        assert getASTNode(ast, "Function", 1).name == "newName"


    def test_doesntBarfOnInheritanceHierarchies(self):
        src = testdata.ClassDefWithCompoundInheritanceHierarchy
        ast = self.runTarget(src, "TheClass.theMethod", "newName")

    def test_renamesMethodWhenMethodCallFromOtherMethodInSameClass(self):
        src = testdata.ClassWithRefToOwnMethod
        ast = self.runTarget(src, "TheClass.theMethod", "newName")
        assert getASTNode(ast, "Getattr", 0).attrname == "newName"

    def test_doesntBarfOnNestedClasses(self):
        src = testdata.ClassInAClass
        ast = self.runTarget(src, "TheClass.theMethod", "newName")

    def test_renamesMethodWhenBaseClassesArentInAST(self):
        src = testdata.BaseClassNotInAST
        ast = self.runTarget(src, "TheClass.theMethod", "newName")
        assert getASTNode(ast, "Function", 0).name == "newName"

class RenameMethodTests_ImportsClass:
    # template method
    def runTarget(self, src, klassfqn, methodname, newname):
        """
        Must provide class in ast at a.b.bah.TheClass
        with method 'theMethod'
        """

    def test_renamesMethodOnDerivedClassInstance(self):
        src = testdata.MethodOverriddenOnDerivedClassInstance
        ast = self.runTarget(src, "a.b.bah.TheClass.theMethod", "newName")
        assert getASTNode(ast, "Function", 0).name == "newName"

#def test_renamesMethodOnBaseClassInstance(self):
#    src = testdata.MethodOverriddenOnDerivedClassInstance
#    ast = self.runTarget(src,"a.b.bah.TheClass","theMethod","newName")
#    assert getASTNode(ast,"Function",0).name == "newName"



class RenameMethodReferenceTests_ImportsClass:

    # template method
    def runTarget(self, src, fqn, newname):
        """
        Must provide class in ast at a.b.bah.TheClass
        with method 'theMethod'
        """

    def test_renamesReferenceOfClassImportedAsAnotherName(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceWithImportAs[1:]
        ast = self.runTarget(src, "a.b.bah.TheClass.theMethod", "newName")
        assert getASTNode(ast, "Getattr", 0).attrname == "newName"

    def test_renamesReferenceWhenObjectCreationAndReferenceInModuleScope(self):
        src = testdata.ObjectCreationAndMethodReference[1:]
        ast = self.runTarget(src, "a.b.bah.TheClass.theMethod", "newName")
        assert getASTNode(ast, "Getattr", 0).attrname == "newName"


    def test_renamesReferenceWhenObjectCreatedInSameFunctionAsReference(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceWithImport[1:]
        ast = self.runTarget(src, "a.b.bah.TheClass.theMethod", "newName")
        assert getASTNode(ast, "Getattr", 2).attrname == "newName"

    def test_doesntrenameDifferentMethodReferenceWhenObjectCreatedInSameScope(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceWithImport[1:]
        ast = self.runTarget(src, "a.b.bah.TheClass.differentMethod", "newName")
        assert astToString(ast) == src

    def test_doesntrenameMethodReferenceWhenDifferentObjectCreatedInSameScope(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceWithImport[1:]
        ast = self.runTarget(src, "a.b.bah.DifferentClass.theMethod", "newName")
        assert astToString(ast) == src

    def test_renamesReferenceOfImportedClass(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceWithImportOfNestedClass[1:]
        ast = self.runTarget(src, "a.b.bah.TheClass.theMethod", "newName")
        assert getASTNode(ast, "Getattr", 2).attrname == "newName"

    def test_doesntRenameReferenceOfDifferentImportedClass(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceButNotImported[1:]
        ast = self.runTarget(src, "a.b.bah.DifferentClass.theMethod", "newName")
        assert astToString(ast) == src

    def test_renamesReferenceOfClassImportedWithFromClause(self):
        src = testdata.functionContaingObjectCreationAndMethodReferenceWithImportFrom[1:]
        ast = self.runTarget(src, "a.b.bah.TheClass.theMethod", "newName")
        assert getASTNode(ast, "Getattr", 0).attrname == "newName"

    def test_doesntrenameReferenceOfClassImportedWithDifferentAsClause(self):
        src = testdata.FunctionContainingMethodReferenceButNotSameAsOneImported[1:]
        ast = self.runTarget(src, "a.b.bah.TheClass.theMethod", "newName")
        assert astToString(ast) == src

    def test_renamesReferenceOfClassImportedWithFromFooImportStar(self):
        src = testdata.FunctionContainingMethodReferenceWithFromImportStar[1:]
        ast = self.runTarget(src, "a.b.bah.TheClass.theMethod", "newName")
        assert getASTNode(ast, "Getattr", 0).attrname == "newName"

    def test_renamesMethodReferenceOfInstanceCreatedInParentScope(self):
        src = testdata.ObjectCreatedInParentScopeToMethodReference[1:]
        ast = self.runTarget(src, "a.b.bah.TheClass.theMethod", "newName")
        assert getASTNode(ast, "Getattr", 0).attrname == "newName"

    def test_doesntRenameMethodWhenObjectCreatedInChildScopeToMethodReference(self):
        src = testdata.ObjectCreatedInChildScopeToMethodReference[1:]
        ast = self.runTarget(src, "a.b.bah.TheClass.theMethod", "newName")
        assert astToString(ast) == src

    def test_renamesReferenceOnDerivedClassInstance(self):
        src = testdata.MethodCalledOnDerivedClassInstance
        ast = self.runTarget(src, "a.b.bah.TheClass.theMethod", "newName")
        assert getASTNode(ast, "Getattr", 2).attrname == "newName"



class RenameMethodReferenceTests:
    # Generic tests. These tests are designed to be run in the context of a ui
    #  and in a package hierarchy structure

    # template method
    def runTarget(self, src, fqn, newname):
        # see concrete subclasses for implementation
        pass

    def test_doesntBarfWhenConfrontedWithComplexReturnTypes(self):
        src = testdata.FunctionReturnsArrayExpr
        ast = self.runTarget(src, "TheClass.theMethod", "newName")

    def test_doesntbarfWhenCallMadeOnInstanceReturnedFromFnCall(self):
        src = testdata.FunctionCallReturnsInstance
        ast = self.runTarget(src, "TheClass.theMethod", "newName")

    def test_doesntStackOverflowOnRecursiveFunctions(self):
        src = testdata.RecursiveFunctionDef
        ast = self.runTarget(src, "TheClass.theMethod", "newName")

    def test_renamesMethodReferenceOfInstanceCreatedInParentScopeAfterFunction(self):
        src = testdata.ObjectCreatedInParentScopeAfterMethodReference
        ast = self.runTarget(src, "TheClass.theMethod", "newName")
        assert getASTNode(ast, "Getattr", 0).attrname == "newName"

    def test_renamesMethodReferenceOfInstanceObtainedByCallingFunction(self):
        src = testdata.ObjectCreatedInFunctionAndUsedByCaller[1:]
        ast = self.runTarget(src, "TheClass.theMethod", "newName")
        assert getASTNode(ast, "Getattr", 0).attrname == "newName"

    def test_renamesMethodReferenceOfInstanceCreatedInAnotherFunction(self):
        src = testdata.ObjectReturnedByFunctionAndUsedByCallerInOtherFunction[1:]
        ast = self.runTarget(src, "TheClass.theMethod", "newName")
        assert getASTNode(ast, "Getattr", 0).attrname == "newName"

    def test_renamesMethodReferenceOfInstanceCreatedInSubsequentFunction(self):
        src = testdata.ObjectReturnedByFunctionAndUsedByCallerInPreviousFunction[1:]
        ast = self.runTarget(src, "TheClass.theMethod", "newName")
        assert getASTNode(ast, "Getattr", 0).attrname == "newName"

    def test_renamesMethodReferenceOnInstanceThatIsAnAttributeOfSelf(self):
        src = testdata.ClassInstanceIsAttributeOfSelf
        ast = self.runTarget(src, "TheClass.theMethod", "newName")
        #print getASTNode(ast, "Getattr", 0)
        assert getASTNode(ast, "Getattr", 0).attrname == "newName"

    def test_doesntBarfOnGetattrThatItCantDeduceTypeOf(self):
        src = testdata.GetattrWithNoWayToDeduceType
        ast = self.runTarget(src, "TheClass.theMethod", "newName")

    def test_renamesMethodInBaseClass(self):
        src = testdata.ClassHierarchy
        ast = self.runTarget(src, "TheClass.theMethod", "newName")
        baseclass = getMatchingASTNode(ast, "Class", name = "root")
        basemethod = getASTNode(baseclass, "Function")
        assert basemethod.name == "newName"

    def test_renamesMethodOnAnotherSubClassOfBaseClass(self):
        src = testdata.ClassHierarchy
        ast = self.runTarget(src, "TheClass.theMethod", "newName")
        baseclass = getMatchingASTNode(ast, "Class", name = "a")
        basemethod = getASTNode(baseclass, "Function")
        assert basemethod.name == "newName"

    def test_renameMethodDoesntBarfOnNoneAsDefaultArgToMethod(self):
        src = testdata.method_with_NoneAsDefaultArg
        ast = self.runTarget(src, "TheClass.theMethod", "newName")


# tests that cover stuff not renamed automatically
# (I.e. are renamed after user manually expresses desire to do so)
class RenameMethodAfterPromptTests:
    def test_renamesReferenceWhenMethodCallDoneOnInstanceCreation(self):
        # i.e.   TheClass().theMethod()
        src = testdata.RefCalledOnClassInstanceCreation
        ast = self.runTarget(src, "TheClass.theMethod", "newName")
        assert getASTNode(ast, "Getattr", 0).attrname == "newName"


    def test_renamesReferenceInMiddleOfBiggerCompoundCall(self):
        src = testdata.MethodRefIsPartOfABiggerCompoundCall
        ast = self.runTarget(src, "TheClass.theMethod", "newName")
        assert getASTNode(ast, "Getattr", 1).attrname == "newName"




def createNestedAST(src1, src2):
    createPackageStructure(src1, src2)
    ast = load(pkgstructureBasedir)
    removePackageStructure()
    return ast


class CreateASTHelper:
    def createAST(self, src):
        ast = Top([], [], [])
        createSourceNodeAt(ast, src, "foo")
        addtypeinfo(ast)
        return ast



class TestRenameMethodWithSingleModule(RenameMethodTests, RenameMethodReferenceTests, unittest.TestCase, CreateASTHelper):
    # template method 
    def runTarget(self, src, fqn, newname):
        ast = self.createAST(src)
        renameMethod(ast, "foo."+fqn, newname)
        return ast


class TestRenameMethodWithDirectoryStructure(RenameMethodTests, RenameMethodReferenceTests, unittest.TestCase):
    # template method
    def runTarget(self, src, fqn, newname):
        ast = createNestedAST("pass", src)
        renameMethod(ast, "a.b.bah."+fqn, newname)
        return getMatchingASTNode(ast, "Module", name = "bah")


class TestRenameMethodReferenceWithDirectoryStructure(RenameMethodTests_ImportsClass, RenameMethodReferenceTests_ImportsClass, unittest.TestCase):
    # template method
    def runTarget(self, src, fqn, newname):
        ast = createNestedAST(src, testdata.TheClass)
        renameMethod(ast, fqn, newname)
        return getMatchingASTNode(ast, "Module", name = "foo")


class TestRenameMethodReferenceTwice(RenameMethodTests_ImportsClass, RenameMethodReferenceTests_ImportsClass, unittest.TestCase):
    # template method
    def runTarget(self, src, fqn, newname):
        ast = createNestedAST(src, testdata.TheClass)
        renameMethod(ast, fqn, "anotherName")
        renameMethod(ast, fqn_rcdr(fqn)+".anotherName", newname)
        return getMatchingASTNode(ast, "Module", name = "foo")

class TestRenameMethodStuffCorrectlyAfterPromptReturnsTrue(unittest.TestCase, 
 RenameMethodAfterPromptTests, 
 CreateASTHelper):

    class MockPromptCallback(Mock):
        def callback(self, node, scopenode):
            return 1

    def runTarget(self, src, fqn, newname):
        ast = self.createAST(src)
        renameMethod(ast, "foo."+fqn, newname, promptCallback = self.MockPromptCallback().callback)
        return ast



class TestCantDeduceTypeCallbackMechanism(unittest.TestCase, CreateASTHelper):



    # need to replace with proper mock object
    class MockCallback(Mock):
        def callback(self, node, scopenode):
            return 1

    def test_callbackNotCalledIfEngineCanDeduceTypes(self):
        src = testdata.ObjectReturnedByFunctionAndUsedByCallerInPreviousFunction
        ast = self.createAST(src)
        mockcb = self.MockCallback()
        renameMethod(ast, "foo.TheClass.theMethod", "newName", promptCallback = mockcb.callback)
        assert len(mockcb.mockCalledMethods) == 0

    def test_callbackNotCalledIfEngineCanDeduceThatInstanceIsNotOfTargetClassType(self):
        src = testdata.ObjectReturnedByFunctionAndUsedByCallerInPreviousFunction
        ast = self.createAST(src)
        mockcb = self.MockCallback()
        renameMethod(ast, "foo.NotTheClass.theMethod", "newName", promptCallback = mockcb.callback)
        assert len(mockcb.mockCalledMethods) == 0

    def test_callbackCalledIfEngineCantDeduceType(self):
        class MockCallback(Mock):  # pythonmock needs to support expectation lists
            def callback(self, node, scopenode):
                assert node.expr.name == "a"
                assert node.attrname == "theMethod"
                assert isinstance(scopenode, compiler.ast.Function)
                return 1
        src = testdata.ObjectRefWithNoDeducableType
        ast = self.createAST(src)
        mockcb = MockCallback()
        renameMethod(ast, "foo.TheClass.theMethod", "newName", promptCallback = mockcb.callback)
        assert len(mockcb.mockCalledMethods["callback"]) == 1


class TestRenameMethodReturnsNumberOfChanges(unittest.TestCase, CreateASTHelper):

    def test_returnsCorrectFiguresIfMethodsWereRenamed(self):
        src = testdata.class_and_method
        ast = self.createAST(src)
        numMeths, numRefs = renameMethod(ast, "foo.a.foo", "bah")
        assert numMeths == 1
        assert numRefs == 0

    def test_returnsCorrectFiguresIfMethodReferencesWereRenamed(self):
        src = testdata.ObjectCreatedInFunctionAndUsedByCaller
        ast = self.createAST(src)
        numMeths, numRefs = renameMethod(ast, "foo.TheClass.theMethod", "bah")
        assert numMeths == 1
        assert numRefs == 1


class TestSubclassMap(unittest.TestCase):
    def test_addsSubClass(self):
        src = testdata.testdata_inheritance
        ast = createSourceNodeAt(Top(), src, "a")
        target = SubclassMap(ast)
        classnode = getASTNode(ast, "Class", 0)
        derivedclass = getASTNode(ast, "Class", 1)
        assert target.getSubClasses(classnode) == [derivedclass]

    def test_returnsEmptyListIfNoSubclasses(self):
        src = testdata.class_and_method
        ast = createSourceNodeAt(Top(), src, "a")
        target = SubclassMap(ast)
        classnode = getASTNode(ast, "Class", 0)
        assert target.getSubClasses(classnode) == []


if __name__ == "__main__":
    unittest.main()
